#include "libcDAG.hpp"
#include <assert.h>
#include "cDAG_Scheduler.hpp"
#include "collective_detection.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <queue>
#include <list>

static inline int _add_element_to_schedule(schedule_t *schedule, int offset, cDAG_Graph graph, int ind) {

  int j;

  SCHED_ADD(&graph->vertexes[ind].type, *schedule, offset, sizeof(char));
  SCHED_ADD(&graph->vertexes[ind].num_rev, *schedule, offset, sizeof(int));
  SCHED_ADD(&graph->vertexes[ind].num_adj, *schedule, offset, sizeof(int));
  for(j = 0; j<graph->vertexes[ind].num_adj; j++) {
    SCHED_ADD(&graph->vertexes[ind].adj[j], *schedule, offset, sizeof(int));
  }

  SCHED_ADD(graph->vertexes[ind].args, *schedule, offset, type_get_size(graph->vertexes[ind].type));

  if (graph->vertexes[ind].type == T_OP) {
	  offset -= sizeof(void*); // the last member of the NBC_Args_op struct is the pointer to the args
	                           // however, the actual args should be encoded here, not the pointer
	  int argsize = ((NBC_Args_op*) (graph->vertexes[ind].args))->argsize;
	  void* realargs = ((NBC_Args_op*) (graph->vertexes[ind].args))->args;
	  SCHED_ADD(realargs, *schedule, offset, argsize);
  }

  return offset;
}

cDAG_Schedule cDAG_Compile(cDAG_Graph graph) {

#ifdef HAVE_PGT
  PGT_Rec(1);  // this records the duration of the graph-building id: 1
#endif

  cDAG_Schedule sched={NULL, NULL, 0};

  schedule_t schedule=NULL;
  int i,j;
  int offset=0; // offset in bytes
  std::queue<int> q;

  /* we store the offsets of the elements (vertexes) in this array as we
   * generate the schedule ... those will be used to replace the real
   * offsets in a second run */
  int *offsets = (int*)malloc(graph->num_vertexes*sizeof(int));
  for(i=0; i<graph->num_vertexes; i++) offsets[i]=-1;

  /* we store the incoming edges of the elements (vertexes) in this array 
   * because we have to modify them in the topological sort and we don't
   * want to destroy the original graph */
  int *inedges = (int*)malloc(graph->num_vertexes*sizeof(int));
  for(i=0; i<graph->num_vertexes; i++) inedges[i]=graph->vertexes[i].num_rev;

  /* this is in fact a topological sort:
   * 1) find all vertices with 0 inedges */
  for(i=0; i<graph->num_vertexes; i++) {
    if(inedges[i] == 0) {
      /* this vertex has 0 dependencies */
      q.push(i); // add to ready-queue
    }
  }

  int qsize = q.size();
  SCHED_ADD(&graph->scratchpad_size, schedule, offset, sizeof(unsigned int));
  SCHED_ADD(&qsize, schedule, offset, sizeof(int));

  /* toposort output */
  while(!q.empty()) {
    int ind = q.front();
    q.pop();
    offsets[ind] = offset; // save offset for this element
    offset = _add_element_to_schedule(&schedule, offset, graph, ind);
    inedges[ind]--; // don't look at this vertex again

    /* for all edges it's pointing to */
    for(j = 0; j<graph->vertexes[ind].num_adj; j++) {
      int adjind = graph->vertexes[ind].adj[j];
      inedges[adjind]--;
      if(inedges[adjind] == 0) {
        q.push(adjind);
      }
    }
  }

  sched.size = offset;

  for(i=0; i<graph->num_vertexes; i++) {
    if(inedges[i] > 0) printf("*** dependency graph has a cycle!\n");
  }

  /* now we need to get the offsets right ... the offsets are just the
   * indices in the graph's vertex array at this point. We need to
   * replace them with the real offsets from offsets[ind] :) .. let's
   * just walk the schedule for this ... */
  for(i=0; i<graph->num_vertexes; i++) {
    offset = offsets[i];

    char type; SCHED_GET(&type, schedule, offset, sizeof(char));
    int numprev; SCHED_GET(&numprev, schedule, offset, sizeof(int));
    int numadj; SCHED_GET(&numadj, schedule, offset, sizeof(int));

    //NBC_DEBUG(10, "offset: %i; found type %i, numprev: %i, numadj: %i, adj:", offsets[i], type, numprev, numadj);
    for(j=0;j<numadj; j++) {
      int adj; SCHED_GET(&adj, schedule, offset, sizeof(int));
      offset -= sizeof(int);
      SCHED_SET(&offsets[adj], schedule, offset, sizeof(int));
      //NBC_DEBUG(10, " %i (%i), ", adj, offsets[adj]);
    }
    //NBC_DEBUG(10, "\n");
  }

  free(offsets);
  free(inedges);
  sched.active = schedule;

#ifdef HAVE_PGT
  PGT_Rec(2);  // this records the duration of the compile-to-binsched phase: 2
#endif

  if (graph->do_coll_detection) perform_collective_detection(&sched);

  // copy just created schedule to schedule copy

  assert(sched.active!=NULL);
  assert(sched.size > 0);
  sched.copy = (char*)malloc(sched.size);
  memcpy(sched.copy, sched.active, sched.size);

  return sched;

}

cDAG_Schedule cDAG_Compile_internal(cDAG_Graph graph) {

  cDAG_Schedule sched={NULL, NULL, 0};

  schedule_t schedule=NULL;
  int i,j;
  int offset=0; // offset in bytes
  std::queue<int> q;

  /* we store the offsets of the elements (vertexes) in this array as we
   * generate the schedule ... those will be used to replace the real
   * offsets in a second run */
  int *offsets = (int*)malloc(graph->num_vertexes*sizeof(int));
  for(i=0; i<graph->num_vertexes; i++) offsets[i]=-1;

  /* we store the incoming edges of the elements (vertexes) in this array 
   * because we have to modify them in the topological sort and we don't
   * want to destroy the original graph */
  int *inedges = (int*)malloc(graph->num_vertexes*sizeof(int));
  for(i=0; i<graph->num_vertexes; i++) inedges[i]=graph->vertexes[i].num_rev;

  /* this is in fact a topological sort:
   * 1) find all vertices with 0 inedges */
  for(i=0; i<graph->num_vertexes; i++) {
    if(inedges[i] == 0) {
      /* this vertex has 0 dependencies */
      q.push(i); // add to ready-queue
    }
  }

  int qsize = q.size();
  SCHED_ADD(&graph->scratchpad_size, schedule, offset, sizeof(unsigned int));
  SCHED_ADD(&qsize, schedule, offset, sizeof(int));

  /* toposort output */
  while(!q.empty()) {
    int ind = q.front();
    q.pop();
    offsets[ind] = offset; // save offset for this element
    offset = _add_element_to_schedule(&schedule, offset, graph, ind);
    inedges[ind]--; // don't look at this vertex again

    /* for all edges it's pointing to */
    for(j = 0; j<graph->vertexes[ind].num_adj; j++) {
      int adjind = graph->vertexes[ind].adj[j];
      inedges[adjind]--;
      if(inedges[adjind] == 0) {
        q.push(adjind);
      }
    }
  }

  sched.size = offset;

  for(i=0; i<graph->num_vertexes; i++) {
    if(inedges[i] > 0) printf("*** dependency graph has a cycle!\n");
  }

  /* now we need to get the offsets right ... the offsets are just the
   * indices in the graph's vertex array at this point. We need to
   * replace them with the real offsets from offsets[ind] :) .. let's
   * just walk the schedule for this ... */
  for(i=0; i<graph->num_vertexes; i++) {
    offset = offsets[i];

    char type; SCHED_GET(&type, schedule, offset, sizeof(char));
    int numprev; SCHED_GET(&numprev, schedule, offset, sizeof(int));
    int numadj; SCHED_GET(&numadj, schedule, offset, sizeof(int));

    //NBC_DEBUG(10, "offset: %i; found type %i, numprev: %i, numadj: %i, adj:", offsets[i], type, numprev, numadj);
    for(j=0;j<numadj; j++) {
      int adj; SCHED_GET(&adj, schedule, offset, sizeof(int));
      offset -= sizeof(int);
      SCHED_SET(&offsets[adj], schedule, offset, sizeof(int));
      //NBC_DEBUG(10, " %i (%i), ", adj, offsets[adj]);
    }
    //NBC_DEBUG(10, "\n");
  }

  free(offsets);
  free(inedges);
  sched.active = schedule;

  return sched;

}

// WTF is wrong with this std::queue??? It takes ages to initialize!!!
std::queue<int> q;

/* returns SCHED_DONE when schedule finished and SCHED_CONTINUE when
 * schedule is still running */
int cDAG_Test(cDAG_Handle handle) {
  int offset = 0;
  int i,j;

  schedule_t schedule = handle->schedule;

#ifdef TEST_LIST
  std::list<std::pair<sched_req,int> > *reqs = handle->reqs; // outstanding requests
#endif
#ifdef TEST_VEC
  std::vector<sched_req> *reqs = handle->reqs; // outstanding requests
  std::vector<int> *offsets = handle->offsets; // offsets for requests
#endif

  int init, scratchpad_size;
  SCHED_GET(&scratchpad_size, schedule, offset, sizeof(int));
  SCHED_GET(&init, schedule, offset, sizeof(int));

  if(handle->state == cDAG_Handle_struct::NEW) {
    for(i=0; i<init;i++) {
      q.push(offset);
      NBC_DEBUG(10, "new handle -- element at offset: %i (%lu) is independent!\n", offset, schedule);
      char type;
      SCHED_GET(&type, schedule, offset, sizeof(char));
      offset += sizeof(int);
      int numadj;
      SCHED_GET(&numadj, schedule, offset, sizeof(int));
      offset += numadj * sizeof(int);
      if ((type == T_SEND) || (type == T_RECV)) offset += type_get_size(type);
      else if (type == T_OP) {
    	  NBC_Args_op op;
    	  SCHED_GET(&op, schedule, offset, type_get_size(T_OP));
    	  offset -= sizeof(void*); // last member of NBC_Args_op is a void pointer, but actually argsize bytes are stored there
    	  offset += op.argsize;
      }

    }
    handle->state = cDAG_Handle_struct::RUNNING;
  }

  while(!q.empty() || reqs->size()) {
    int offset;
    char type; 
    int numprev;
    int numadj;    
    
    if(!q.empty()) {
      offset=q.front();
      q.pop();

      SCHED_GET(&type, schedule, offset, sizeof(char));
      SCHED_GET(&numprev, schedule, offset, sizeof(int));
      SCHED_GET(&numadj, schedule, offset, sizeof(int));
      offset += numadj*sizeof(int);

      NBC_DEBUG(10, "fire offset: %i; found type %i, numprev: %i, numadj: %i \n", offset-2*sizeof(int)-sizeof(char)-numadj*sizeof(int), type, numprev, numadj);

      sched_req req;
      if(execute_element(type, schedule+offset, &req, handle) == SCHED_CONTINUE) {
#ifdef TEST_LIST
        reqs->push_back(std::make_pair(req, offset-2*sizeof(int)-sizeof(char)-numadj*sizeof(int)));
#endif
#ifdef TEST_VEC
       if ((type == T_SEND) || (type == T_RECV)) {
        reqs->push_back(req);
        offsets->push_back(offset-2*sizeof(int)-sizeof(char)-numadj*sizeof(int));
       }
#endif
      }
    } else { /* nothing to do - wait */
      // we don't do the testany anoymore because copying the requests
      // together costs too much time! Thus, we iterate over a list of
      // request, offset pairs!
      offset = SCHED_NONE;
#ifdef TEST_LIST
      for(std::list<std::pair<sched_req,int> >::iterator i=handle->reqs->begin(); i!=handle->reqs->end(); ++i) {
        if(SCHED_DONE == scheduler_test((*i).first)) {
          offset = (*i).second;
          handle->reqs->erase(i);
          break;
        }
      }
      // nothing is finished ... give control back to user
      if (offset == SCHED_NONE) return SCHED_CONTINUE;
#endif
#ifdef TEST_VEC
      int req = scheduler_waitany(&reqs->front(), reqs->size());
      //int req = scheduler_testany(&reqs->front(), reqs->size());
      // nothing is finished ... give control back to user
      if (req == SCHED_NONE) return SCHED_CONTINUE;
      offset = (*offsets)[req];
      offsets->erase(offsets->begin()+req);
      reqs->erase(reqs->begin()+req);
#endif

      offset += sizeof(int)+sizeof(char);
      SCHED_GET(&numadj, schedule, offset, sizeof(int));
      NBC_DEBUG(10, "waited for %i reqs finished number %i with offset: %i and numadj: %i\n", reqs->size()+1, req, offset, numadj);

      for(j=0;j<numadj; j++) {
        int adj; SCHED_GET(&adj, schedule, offset, sizeof(int));
        NBC_DEBUG(10, "offset %i is adjacent\n", adj);
        /* get the number of incoming edges of this adjacent element */
        adj += sizeof(char);
        int adjnumprev; 
        SCHED_GET(&adjnumprev, schedule, adj, sizeof(int));
        adjnumprev--;
        adj -= sizeof(int);
        SCHED_SET(&adjnumprev, schedule, adj, sizeof(int));
        if(adjnumprev == 0) {
          adj -= (sizeof(char) + sizeof(int));
          q.push(adj); 
        }
        NBC_DEBUG(10, "offset: %i - adjnumprev: %i\n", adj, adjnumprev);
      }
    }
  }

  return SCHED_DONE;
}

int cDAG_Wait(cDAG_Handle handle) {
  while(SCHED_DONE != cDAG_Test(handle)) {};

  MPI_Barrier(G_cDAG_WorldComm);
#ifdef TEST_LIST
	delete handle->reqs;
#endif
#ifdef TEST_VEC
	delete handle->reqs;
	delete handle->offsets;
#endif

  free(handle);
#ifdef HAVE_PGT
  PGT_Rec(11); //record end of execution id:11
#endif

  return cDAG_SUCCESS;
}

//void cDAG_Attach(cDAG_Schedule sched, MPI_Comm comm, cDAG_Handle *handle) {
cDAG_Handle cDAG_Run(cDAG_Schedule sched) {
  cDAG_Handle handle = (cDAG_Handle_struct*)malloc(sizeof(cDAG_Handle_struct));

  handle->tmpbuf = NULL;
  handle->state = cDAG_Handle_struct::NEW;
  //handle->comm = comm;
  handle->schedule = sched.active;


#ifdef TEST_LIST
  handle->reqs = new std::list<std::pair<sched_req, int> >;
#endif
#ifdef TEST_VEC
  handle->reqs = new std::vector<sched_req>;
  handle->offsets = new std::vector<int>;
#endif

  handle->tag = G_cDAG_Tag;
  //NBC_DEBUG(3, "got tag %i\n", (*handle)->tag);


  return handle;
}

void cDAG_ReuseSchedule(cDAG_Schedule sched) {
  // memory is already allocated and remains so :)
  memcpy(sched.active, sched.copy, sched.size);
}

